subroutine htrns(tmp,u,v,sch,dx,dz,dxs,dzs,asp, &
     tsurf,tbot,qbot,vboti,dtin,courant)
!***********************************************************************
!*     solve energy equation by implicit, power-law scheme
!*     adopted boundary conditions are;
!*        fixed temperature on top/bottom boundaries
!*        zero heat flux across side walls
!*
!*     input
!*        tmp: temperature at time step n
!*        u,v: velocity
!*        sch: heat generation
!*        tsurf: surface temperature
!*        tbot: bottom temperature at time step n
!*        vboti: inverse of volume of heat bath at bottom
!*        dx,dz: mesh size
!*        dt: time step
!*     output
!*        tmp: temperature at time step n+1
!*        tbot: bottom temperature at time step n+1
!*        qbot: bottom heat flux
!*     local
!*        pe: Peclet number
!*        twk: work array for temperature
!*        fx,fz: heat flux in x/z direction
!**********************************************************************
  implicit none
  include "meshnum.h"
!c
  integer :: i,j,ij
  real(kind=8),dimension(iax,jax),intent(in) :: u,v,sch
  real(kind=8),dimension(iax,jax),intent(inout) :: tmp
  real(kind=8),dimension(iax),intent(in) :: dx,dxs
  real(kind=8),dimension(jax),intent(in) :: dz,dzs
  real(kind=8),intent(in) :: asp,dtin,courant,tsurf,vboti
  real(kind=8),intent(inout) :: tbot,qbot
  real(kind=8),dimension(iax,jax) :: twk,difx,difz
  real(kind=8),dimension(iax,jax) :: cax,caz,cbx,cbz
  real(kind=8) :: tb,dt,pe,faca,facb,dum
  real(kind=8),dimension(iax,jax,5) :: at1
  real(kind=8),dimension(iax,jax) :: bt1
!c
  ! variables for LAPACK subroutines (dgbtrf and dgbtrs)
  ! nm: dimension of array
  ! kl: lower band width
  ! ku: upper band width
  ! ldab: leading dimension of storage (.ge.2*kl+ku+1)
  ! ab(lda,nm): coefficient band matrix
  ! b(nm): RHS vector
  ! ipiv(nm) : for pivoting
  ! info
  integer,parameter :: m1=jax-2,kl=m1,ku=m1
  integer,parameter :: nm=(iax-2)*(jax-2),ldab=2*kl+ku+1
  real(kind=8),dimension(ldab,nm) :: ab
  integer,dimension(nm) :: ipiv
  real(kind=8),dimension(nm) :: bx
  integer :: info
!c
  dt = dtin*courant
!c
!c ---------------------------------------
!c     temperature change in heat bath
!c ---------------------------------------
!c
  tb = 0.d0
  do i = 2,iax-1
     tb = tb + tmp(i,2)*dx(i)
  enddo
  tb = tb/asp
  qbot = 2.d0*(tbot-tb)/dz(2)
  tbot = tbot - vboti*qbot*dt
!c
!c -----------------------
!c     set work arrays
!c -----------------------
!c
  do j = 2,jax-1
     do i = 2,iax-1
        twk(i,j) = tmp(i,j)
     enddo
  enddo
!c
!c     top/bottom boundary
!c
  do i = 2,iax-1
     twk(i,1) = (2.d0*dzs(2)*tbot-dz(1)*tmp(i,2))/dz(2)
     twk(i,jax) = (2.d0*dzs(jax)*tsurf-dz(jax)*tmp(i,jax-1)) &
          /dz(jax-1)
  enddo
!c
!c     side walls
!c
  do j = 1,jax
     twk(1,j) = tmp(2,j)
     twk(iax,j) = tmp(iax-1,j)
  enddo
!c
!c -------------------
!c     conductance
!c -------------------
!c
  do j = 2,jax-1
     do i = 2,iax
        difx(i,j) = 1.d0/dxs(i)
     enddo
  enddo
!c
  do j = 2,jax
     do i = 2,iax-1
        difz(i,j) = 1.d0/dzs(j)
     enddo
  enddo
!c
!c ----------------------------------------
!c     calculate heat flux coefficients
!c ----------------------------------------
!c
  do j = 2,jax-1
     do i = 2,iax
        pe = u(i,j)/difx(i,j)
        dum = 1.d0-0.1d0*dabs(pe)
        faca = dmax1(0.d0,dum**5)+dmax1(0.d0,-pe)
        facb = faca + pe
        cax(i,j) = difx(i,j)*faca
        cbx(i,j) = difx(i,j)*facb
     enddo
  enddo
!c
  do j = 2,jax
     do i = 2,iax-1
        pe = v(i,j)/difz(i,j)
        dum = 1.d0-0.1d0*dabs(pe)
        faca = dmax1(0.d0,dum**5)+dmax1(0.d0,-pe)
        facb = faca + pe
        caz(i,j) = difz(i,j)*faca
        cbz(i,j) = difz(i,j)*facb
     enddo
  enddo
!c
!c ----------------------------------------------
!c     calculate coefficients for temperature
!c ----------------------------------------------
!c
  at1(:,:,:) = 0.d0
  bt1(:,:) = 0.d0
  do j = 2,jax-1
     do i = 2,iax-1
        at1(i,j,1) = -cbx(i,j)*dt/dx(i)
        at1(i,j,2) = -cbz(i,j)*dt/dz(j)
        at1(i,j,4) = -caz(i,j+1)*dt/dz(j)
        at1(i,j,5) = -cax(i+1,j)*dt/dx(i)
        at1(i,j,3) = 1.d0 + dt*( &
             (cax(i,j)+cbx(i+1,j))/dx(i) &
             +(caz(i,j)+cbz(i,j+1))/dz(j))
        bt1(i,j) = twk(i,j) + sch(i,j)*dt
     enddo
  enddo
!c
!c ---------------------------
!c     boundary conditions
!c ---------------------------
!c
  do i = 2,iax-1
     at1(i,2,3) = at1(i,2,3) -     dz(1)/dz(2)*at1(i,2,2)
     bt1(i,2)   = bt1(i,2) - 2.d0*dzs(2)/dz(2)*at1(i,2,2)*tbot
     at1(i,2,2) = 0.d0
  enddo
  do i = 2,iax-1
     at1(i,jax-1,3) = at1(i,jax-1,3) -    dz(jax)/dz(jax-1)*at1(i,jax-1,4)
     bt1(i,jax-1)   = bt1(i,jax-1) -2.d0*dzs(jax)/dz(jax-1)*at1(i,jax-1,4)*tsurf
     at1(i,jax-1,4) = 0.d0
  enddo
  do j = 2,jax-1
     at1(2,j,3) = at1(2,j,3) + at1(2,j,1)
     at1(2,j,1) = 0.d0
  enddo
  do j = 2,jax-1
     at1(iax-1,j,3) = at1(iax-1,j,3) + at1(iax-1,j,5)
     at1(iax-1,j,5) = 0.d0
  enddo
!c
!c /////////////////////////////////////
!c     begin linear solver
!c /////////////////////////////////////
  ab(:,:) = 0.d0
  do j = 2,jax-1
     do i = 2,iax-1
        ij = j-1+(jax-2)*(i-2)
        ab(kl+ku+1-m1,ij) = at1(i-1,j  ,5) ! (i-1,j  )
        ab(kl+ku+1-1 ,ij) = at1(i  ,j-1,4) ! (i  ,j-1)
        ab(kl+ku+1   ,ij) = at1(i  ,j  ,3) ! (i  ,j  )
        ab(kl+ku+1+1 ,ij) = at1(i  ,j+1,2) ! (i  ,j+1)
        ab(kl+ku+1+m1,ij) = at1(i+1,j  ,1) ! (i+1,j  )
        bx(ij) = bt1(i,j)
     enddo
  enddo
  !
  call dgbtrf(nm,nm,kl,ku,ab,ldab,ipiv,info)
  if (info /= 0) then
     write(*,*)"LU factorization for general band matrix failed",info
     stop
  end if
  call dgbtrs("N",nm,kl,ku,1,ab,ldab,ipiv,bx,nm,info)
  !
  do i = 2,iax-1
     do j = 2,jax-1
        ij = j-1+(jax-2)*(i-2)
        twk(i,j) = bx(ij)
     enddo
  enddo
!c
!c     top/bottom boundary
!c
  do i = 2,iax-1
     twk(i,1) = (2.d0*dzs(2)*tbot-dz(1)*twk(i,2))/dz(2)
     twk(i,jax) = (2.d0*dzs(jax)*tsurf-dz(jax)*twk(i,jax-1)) &
          /dz(jax-1)
  enddo
!c
!c     side walls
!c
  do j = 1,jax
     twk(1,j) = twk(2,j)
     twk(iax,j) = twk(iax-1,j)
  enddo
!c
  do j = 1,jax
     do i = 1,iax
        tmp(i,j) = twk(i,j)
     enddo
  enddo
!c
  return
end subroutine htrns
